/***************************************************************************
$name: ls.c 
$version: 0.8.0
$date_modified: 24. january 2003
$description: ls displays the contents of a directory similar it is done
              on UNIX. This is one of the binaries needed for the UNIX
	      environment on NetWare.
$owner: Ulrich Neumann, Germany

BSD License
MAKE SURE THAT YOUR NLM IS MARKED PSEUDOPREEMPTION AND MULTIPLE !
****************************************************************************/

#include <stdio.h>
#include <screen.h>
#include <dirent.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <getopt.h>

/* Return an int indicating the result of comparing two integers.
   Subtracting doesn't always work, due to overflow.  */
#define longdiff(a, b) ((a) < (b) ? -1 : (a) > (b))

# define ST_TIME_CMP_NS(a, b, ns) 0
#define ST_TIME_CMP(a, b, s, ns) \
  ((a).s < (b).s ? -1 : (a).s > (b).s ? 1 : ST_TIME_CMP_NS(a, b, ns))
#define ATIME_CMP(a, b) ST_TIME_CMP (a, b, st_atime.tv_sec, st_atim.ST_MTIM_NSEC)
#define CTIME_CMP(a, b) ST_TIME_CMP (a, b, st_ctime.tv_sec, st_ctim.ST_MTIM_NSEC)
#define MTIME_CMP(a, b) ST_TIME_CMP (a, b, st_mtime.tv_sec, st_mtim.ST_MTIM_NSEC)

char *getbsize(int *headerlenp, long *blocksizep);

int		maxnamelen = 0;
int		lastnamepos = 0;
long		blocksize;		/* block size units		*/
int		termwidth = 80;		/* default terminal width	*/
extern	int	optind;
char		*program_name;

/* The table of files in the current directory:

   `files' points to a vector of `struct fileinfo', one per file.
   `nfiles' is the number of elements space has been allocated for.
   `files_index' is the number actually in use.  */

/* Address of block containing the files that are described.  */
static struct fileinfo *files;  /* FIXME: rename this to e.g. cwd_file */

/* Length of block that `files' points to, measured in files.  */
static int nfiles;  /* FIXME: rename this to e.g. cwd_n_alloc */

/* Index of first unused in `files'.  */
static int files_index;  /* FIXME: rename this to e.g. cwd_n_used */

enum filetype
{
	unknown,
	fifo,
	chardev,
	directory,
	blockdev,
	normal,
	symbolic_link,
	sock,
	arg_directory,
};

struct fileinfo
{
	/* The file name. */
	char *name;

	struct stat stat;

	/* For symbolic link, name of the file linked to, otherwise zero. */
	char *linkname;

	/* For symbolic link and long listing, st_mode of file linked to, otherwise
	   zero. */
	unsigned int linkmode;

	/* For symbolic link and color printing, 1 if linked-to file
	   exists, otherwise 0.  */
	int linkok;

	enum filetype filetype;
};

static struct option const long_options[] =
{
	{"almost-all", no_argument, 0, 'A'},
	{"all", no_argument, 0, 'a'},
	{"size", no_argument, 0, 's'},
	{"long", no_argument, 0, 'l'},
	{"reverse", no_argument, 0, 'r'},
	{"time", no_argument, 0, 't'},
	{"ctime", no_argument, 0, 'c'},
	{"accesstime", no_argument, 0, 'u'},
	{"help", no_argument, 0, 'h'},
	{"version", no_argument, 0, 'v'},
	{NULL, 0, NULL, 0}
};

/* flags */
static int f_nosort;		/* don't sort output */
static int f_reversesort;	/* reverse whatever sort is used */
static int f_size;		/* list size in short listing */
static int f_accesstime;	/* use time of last access */
static int f_statustime;	/* use time of last mode change */
static int f_timesort;		/* sort by time vice name */
static int f_listdot;		/* list files beginning with . */
static int f_listdot_implied;	/* list implied files beginning with . */
static int f_longform;		/* long listing format */

static int f_singlecol;		/* use single column output */
static int f_kblocks;		/* print size in kilobytes */

/* prototypes */
static int compare_atime (const struct fileinfo *file1, const struct fileinfo *file2);
static int rev_cmp_atime (const struct fileinfo *file2, const struct fileinfo *file1);
static int compare_ctime (const struct fileinfo *file1, const struct fileinfo *file2);
static int rev_cmp_ctime (const struct fileinfo *file2, const struct fileinfo *file1);
static int compare_mtime (const struct fileinfo *file1, const struct fileinfo *file2);
static int rev_cmp_mtime (const struct fileinfo *file2, const struct fileinfo *file1);
static int compare_size ( const struct fileinfo *file1, const struct fileinfo *file2 );
static int rev_cmp_size ( const struct fileinfo *file2, const struct fileinfo *file1 );
static int compare_name ( const struct fileinfo *file1, const struct fileinfo *file2 );
static int rev_cmp_name ( const struct fileinfo *file2, const struct fileinfo *file1 );

static void	traverse	( int argc, char *argv[] );
static void	sort_files	( void );
void		listdir_long	( void );
void		listdir_short (void);
static void	usage		( void );
static void	version		( void );

int main (int argc, char **argv)
{
	int		ch,
			notused;
	char		*last_param;
	static char	dot[] = ".", *dotav[] = {dot, NULL};

	program_name	= argv[0];

	printf ("argc is: %d, %s\n", argc, argv[argc - 1]);

 	while ((ch = getopt_long(argc, argv, "Aachlrstuv", long_options, NULL)) != -1)
	{
		switch (ch)
		{
			/*
			 * The -1, -C, -x and -l options all override each other so
			 * shell aliasing works right.
			 */
			case 'A':
				f_listdot_implied = 1;
				break;
			case 'a':
				f_listdot = 1;
				break;
			case 's':
				f_size = 1;
				break;
			case 'r':
				f_reversesort = 1;
				break;
			case 't':
				f_timesort = 1;
				break;
			/* The -c and -u options override each other. */
			case 'c':
				f_statustime = 1;
				f_accesstime = 0;
				break;
			case 'l':
				f_longform = 1;
				f_singlecol = 0;
				break;
			case 'u':
				f_accesstime = 1;
				f_statustime = 0;
				break;
			case 'h':
				usage();
				break;
			case 'v':
				version();
				break;

			default:
			case '?':
				usage();
		}
	}

	argc -= optind;
	argv += optind;

	/* If -l or -s, figure out block size. */
	if (f_longform || f_size) {
		if (f_kblocks)
			blocksize = 2;
		else {
			(void)getbsize(&notused, &blocksize);
			blocksize /= 512;
		}
	}

	if (argc)
		traverse(argc, argv);
	else
		traverse(1, dotav);

	return 0;
}

static void
traverse (int argc, char *argv[])
{
	int		len;
	DIR		*d;
	struct dirent	*f;
	int		compare_argv;
	int		add_entry;
	int		i;

	/* 
	 * Build a list of all entries. 
	 */
	compare_argv	= 0;
	add_entry	= 1;
	nfiles		= 100;
	files_index	= 0;
	files		= (struct fileinfo *) malloc (sizeof (struct fileinfo) * nfiles);

	if (argc > 1)
	printf ("d = opensir(\"%s\"", argv[0]);


	if (argc > 0)
		compare_argv = 1;
	if (!(d = opendir(argv[0])))
		d = opendir(".");

	while (f = readdir(d))
	{
		len = strlen(f->d_name) + 1;

		if ( (strncmp (f->d_name, ".", 1) != NULL)
		     || f_listdot 
		     || (((!strcmp(f->d_name, ".") || !strcmp(f->d_name, "..")) && f_listdot_implied)) )
		{
			/* check for the argv[1 - n] entries */
			if (compare_argv)
			{
				add_entry = 0;
				for (i = 0; i < argc; i++)
					if (!strcmp(f->d_name, argv[i]))
						add_entry = 1;
			}

			if (add_entry)
			{
				if (len > maxnamelen)
					maxnamelen = len;

				if (files_index == nfiles)
				{
					nfiles *= 2;
					files = (struct fileinfo *) realloc ((char *) files,
						sizeof (*files) * nfiles);
				}

				if (!( (char *) (files[files_index].name) = (char *) malloc(len) ))
				{
					fprintf(stderr, "Cannot malloc\n");
					exit(1);
				}

				/* copy entry name to file structure */
				strcpy( (char *) (files[files_index].name), f->d_name);

				/* get file stat for file structure */
				if (stat (f->d_name, &files[files_index].stat))
				{
					fprintf(stderr, "stat error\n");
				}

				files_index++;
			}
		}
	}

	if (closedir(d))
	{
		fprintf(stderr, "Could not close directory.\n");
		exit(1);
	}

	/*
	 * Sort the list dependend on several parameters
	 */
	sort_files ();

	/*
	 * print the list
	 */
	if (f_longform)
		listdir_long();
	else	
		listdir_short();

	free (files);
	return;
}

/* Sort the files now in the table.  */

static void
sort_files (void)
{
	int (*sortfunc) ();

	/* Select a sort function. */
	if (f_reversesort) {
		if (f_timesort)
			sortfunc = rev_cmp_ctime;
		else if (f_accesstime && f_timesort)
			sortfunc = rev_cmp_atime;
		else if (f_statustime && f_timesort)
			sortfunc = rev_cmp_mtime;
		else if (f_size)
			sortfunc = compare_size;
		else		/* standard sort is by name. */
			sortfunc = rev_cmp_name;
	} else {
		if (f_timesort)
			sortfunc = compare_ctime;
		else if (f_accesstime && f_timesort)
			sortfunc = compare_atime;
		else if (f_statustime && f_timesort)
			sortfunc = compare_mtime;
		else if (f_size)
			sortfunc = rev_cmp_size;
		else		/* standard sort is by name. */
			sortfunc = compare_name;
	}

	qsort (files, files_index, sizeof (struct fileinfo), sortfunc);
}



static int
compare_ctime (const struct fileinfo *file1, const struct fileinfo *file2)
{
  int diff = CTIME_CMP (file2->stat, file1->stat);
  if (diff == 0)
    diff = strcoll (file1->name, file2->name);
  return diff;
}

static int
rev_cmp_ctime (const struct fileinfo *file2, const struct fileinfo *file1)
{
  int diff = CTIME_CMP (file2->stat, file1->stat);
  if (diff == 0)
    diff = strcoll (file1->name, file2->name);
  return diff;
}

static int
compare_mtime (const struct fileinfo *file1, const struct fileinfo *file2)
{
  int diff = MTIME_CMP (file2->stat, file1->stat);
  if (diff == 0)
    diff = strcoll (file1->name, file2->name);
  return diff;
}

static int
rev_cmp_mtime (const struct fileinfo *file2, const struct fileinfo *file1)
{
  int diff = MTIME_CMP (file2->stat, file1->stat);
  if (diff == 0)
    diff = strcoll (file1->name, file2->name);
  return diff;
}

static int
compare_atime (const struct fileinfo *file1, const struct fileinfo *file2)
{
  int diff = ATIME_CMP (file2->stat, file1->stat);
  if (diff == 0)
    diff = strcoll (file1->name, file2->name);
  return diff;
}

static int
rev_cmp_atime (const struct fileinfo *file2, const struct fileinfo *file1)
{
  int diff = ATIME_CMP (file2->stat, file1->stat);
  if (diff == 0)
    diff = strcoll (file1->name, file2->name);
  return diff;
}

static int
compare_size (const struct fileinfo *file1, const struct fileinfo *file2)
{
	int diff = longdiff (file2->stat.st_size, file1->stat.st_size);
	if (diff == 0)
		diff = strcoll (file1->name, file2->name);
	return diff;
}

static int
rev_cmp_size (const struct fileinfo *file2, const struct fileinfo *file1)
{
	int diff = longdiff (file2->stat.st_size, file1->stat.st_size);
	if (diff == 0)
		diff = strcoll (file1->name, file2->name);
	return diff;
}

static int
compare_name (const struct fileinfo *file1, const struct fileinfo *file2)
{
	return strcoll (file1->name, file2->name);
}

static int
rev_cmp_name (const struct fileinfo *file2, const struct fileinfo *file1)
{
	return strcoll (file1->name, file2->name);
}

void listdir_long (void)
{
	int	file;
	int	f_executable;
	int	f_len;
	time_t	filetime;
	char	filedate[50];
	
	printf ("\r");

	for (file = 0; file < files_index; file++)
	{
		switch (files[file].stat.st_mode & S_IFMT)
		{
			case S_IFLNK:
				putchar ('l');
				break;
			case S_IFREG:
				putchar ('-');
				break;
			case S_IFDIR:
				putchar ('d');
				break;
			case S_IFCHR:
				putchar ('c');
				break;
			case S_IFBLK:
				putchar ('b');
				break;
			case S_IFIFO:
				putchar ('p');
				break;
			case S_IFSOCK:
				putchar ('s');
				break;
			default:
				putchar ('-');
				break;
		}

		/* print UNIX style permissions */
		f_executable = 0;
		f_len = strlen(files[file].name);
		if (f_len > 3)
		{
			if (!strncasecmp(files[file].name + f_len - 4, ".nlm", 4))
				f_executable = 1;
			else if (!strncasecmp(files[file].name + f_len - 4, ".xlm", 4))
				f_executable = 1;
			else if (!strncasecmp(files[file].name + f_len - 4, ".ncf", 4))
				f_executable = 1;
			else if (!strncasecmp(files[file].name + f_len - 3, ".sh", 3))
				f_executable = 1;
			else if (!strncasecmp(files[file].name + f_len - 3, ".pl", 3))
				f_executable = 1;
			else if (!strncasecmp(files[file].name + f_len - 4, ".php", 4))
				f_executable = 1;
			else if (!strncasecmp(files[file].name + f_len - 4, ".cdm", 4))
				f_executable = 1;
			else if (!strncasecmp(files[file].name + f_len - 4, ".ham", 4))
				f_executable = 1;
			else if (!strncasecmp(files[file].name + f_len - 4, ".nss", 4))
				f_executable = 1;
		}
	
		printf ("rw");
		if (f_executable)
			printf ("x");
		else
			printf ("-");
			
		printf ("r-");
		if (f_executable)
			printf ("x");
		else
			printf ("-");

		printf ("r-");
		if (f_executable)
			printf ("x");
		else
			printf ("-");

		/* print user and group ID */
//		printf (" %8d", files[file].stat.st_uid);
//		printf (" %8d", files[file].stat.st_gid);

		/* print file size */
		printf (" %10d", files[file].stat.st_size);

		if (f_accesstime)
			filetime = files[file].stat.st_atime.tv_sec;
		else
			filetime = files[file].stat.st_mtime.tv_sec;
			
		sprintf (filedate, "%s", ctime(&filetime));
		memset (filedate + strlen(filedate) - 1, 0, 1);
		printf (" %s", filedate);

		printf ("  %s\n", files[file].name);
		free(files[file].name);
	}	
	printf ("Total entries: %d\n", files_index);
}


void listdir_short (void)
{
	int	file = 0;
	int	row = 0;
	int	i = 0;
	int	len = 0;

	printf ("\r");

	maxnamelen += 2;	/* increase maxnamelen to make sure that every directory will be printed correct */

	for (file = 0; file < files_index; file++)
	{
		if ((files[file].stat.st_mode & S_IFMT) == S_IFDIR)
			printf ("[");

		printf("%s", files[file].name);
		if ((files[file].stat.st_mode & S_IFMT) == S_IFDIR)
			printf ("]");

		len = strlen(files[file].name);

		if ((files[file].stat.st_mode & S_IFMT) == S_IFDIR)
			len +=2;

		lastnamepos += (maxnamelen + 3);

		if (lastnamepos <= (80 - maxnamelen))
		{
			for (i = 0; i <= (maxnamelen - len) ; i++)
				putchar(' ');
		}
		else
		{
			printf("\n");
			lastnamepos = 0;
		}

		free(files[file].name);
	}
	printf ("\nTotal entries: %d\n", files_index);
}

static void
usage (void)
{
	printf ("Usage: %s [OPTION]... [FILE]...\n\n", program_name);

	printf ("\
List information about the FILEs (the current directory by default).\n\
Sort entries alphabetically if none of -ctu.\n\
\n\
  -a, --all                  do not hide entries starting with .\n\
  -A, --almost-all           do not list implied . and ..\n\
  -c, --ctime                with -lt: sort by, and show, ctime (time of last\n\
                               modification of file status information)\n\
                               with -l: show ctime and sort by name\n\
                               otherwise: sort by ctime\n\
  -l, --long                 use a long listing format\n\
  -r, --reverse              reverse order while sorting\n\
  -s, --size                 sort by file size\n\
  -t, --time                 sort by modification time\n\
  -u, --accesstime           with -lt: sort by, and show, access time\n\
                               with -l: show access time and sort by name\n\
                               otherwise: sort by access time\n\
  -h, --help                 display this help and exit\n\
  -v, --version              output version information and exit\n\
\n\
  ");
	exit(1);
}

static void
version (void)
{
	printf ("LS for NetWare. Version 0.1.0 by Ulrich Neumann, 20. January 2002.\n");
	exit(1);
}
